####################
# File name: A3.py
# This is the programming portion of Assignment 3
# for CPSC 447/547 Intro to Quantum Computing
#
# Name: 
# NetID:
####################

import requirement_A3
import math
import numpy as np


####################
# Place for your own helper functions, if any
####################



####################
# Building Blocks
####################
        
class Qubit(object):
    """Qubit object"""
    def __init__(self, arg, label='q'):
        super(Qubit, self).__init__()
        self.arg = arg
        self.label = label


class QuantumRegister(object):
    """QuantumRegister is where we keep track of qubits"""
    def __init__(self, qubits, label='qreg'):
        super(QuantumRegister, self).__init__()
        if isinstance(qubits, list):
            if ((len(qubits)==0) or (not isinstance(qubits[0], Qubit))):
                raise Exception("Invalid inputs to QuantumRegister.")
            self.size = len(qubits)
            self.label = qubits[0].label
            self.array = qubits
            # new_q is a list of indicator for whether the qubits in this register
            # is newly initialized or copied from other existing registers.
            self.new_q = [False for _ in qubits]
        elif isinstance(qubits, int):
            # New qubits in register
            self.size = qubits
            self.label = label
            self.array = [Qubit(i, label) for i in range(qubits)]
            # new_q is a list of indicator for whether the qubits in this register
            # is newly initialized or copied from other existing registers.
            self.new_q = [True for _ in range(qubits)]
        else:
            raise Exception("Invalid inputs to QuantumRegister.")


class ClassicalRegister(object):
    """ClassicalRegister is where we keep track of measurement outcomes"""
    def __init__(self, num_c, label='creg'):
        super(ClassicalRegister, self).__init__()
        self.size = num_c
        self.label = label
        self.state = np.array([bool(0) for _ in range(num_c)])


class Gate(object):
    """Gate object to describe its name, kind, and matrix"""
    def __init__(self, name, num_q, matrix):
        super(Gate, self).__init__()
        self.name = name 
        self.num_q = num_q
        self.matrix = matrix
    def __repr__(self):
        return self.name

		
####################
# In-House Quantum Circuit
####################

class QuantumCircuit(object):
    """QuantumCircuit"""
    def __init__(self, qubits, cbits, name='module'):
        super(QuantumCircuit, self).__init__()
        self.name = name
        if isinstance(qubits, QuantumRegister):
            if ((qubits.size==0) or (not isinstance(qubits.array[0], Qubit))):
                raise Exception("Invalid inputs to QuantumCircuit.")
            self.num_q = qubits.size
            self.qubits = qubits # use the input qubits
        elif isinstance(qubits, int):
            self.num_q = qubits
            self.qubits = QuantumRegister(qubits) # initialized qubits
        else:
            raise Exception("Invalid inputs to QuantumCircuit.")
        if isinstance(cbits, ClassicalRegister):
            self.num_c = cbits.size
            self.cbits = cbits # use the input cbits
        elif isinstance(cbits, int):
            self.num_c = cbits
            self.cbits = ClassicalRegister(cbits) # initialized cbits
        else:
            raise Exception("Invalid inputs to QuantumCircuit.")
        self.circuit = [] # sequence of instructions

    def _append(self, operation, q_array, c_array):
        # Add new instruction to circuit
        instruction = [operation, q_array, c_array]
        self.circuit.append(instruction)

    def __repr__(self):
        # For displaying quantum circuit
        qasm = ['\n======<CPSC 447/547 QASM>======']
        qasm += ['Qreg: %d, Creg: %d' % (self.num_q, self.num_c)]
        for inst in self.circuit:
            #print(inst)
            (op, q_arr, c_arr) = inst
            inst_str = '%s ' % op.name 
            for q in q_arr:
                qubit = self.qubits.array[q]
                inst_str += '%s%d ' % (qubit.label, qubit.arg)
            inst_str += ', '
            for c in c_arr:
                inst_str += '%s%d ' % (self.cbits.label, c)
            qasm.append(inst_str)
        qasm.append('===============================\n')
        return "\n".join(qasm)

    # Define instruction set
    # Hadamard gate
    def h(self, qubit):
        # Define Gate by its name, kind (number of qubit), and matrix
        HGate = Gate('h', 1, None)
        # add to the end of the self.circuit list
        self._append(HGate, [qubit], [])
        return
    # Pauli X gate
    def x(self, qubit):
        XGate = Gate('x', 1, None)
        self._append(XGate, [qubit], [])
        return
    # Pauli Y gate
    def y(self, qubit):
        YGate = Gate('y', 1, None)
        self._append(YGate, [qubit], [])
        return
    # Pauli Z gate
    def z(self, qubit):
        ZGate = Gate('z', 1, None)
        self._append(ZGate, [qubit], [])
        return
    # Phase gate (sqrt(Z))
    def s(self, qubit):
        SGate = Gate('s', 1, None)
        self._append(SGate, [qubit], [])
        return
    # S dagger gate (adjoint of S gate)
    def sdg(self, qubit):
        SDGGate = Gate('sdg', 1, None)
        self._append(SDGGate, [qubit], [])
        return
    # T gate (sqrt(S))
    def t(self, qubit):
        TGate = Gate('t', 1, None)
        self._append(TGate, [qubit], [])
        return
    # T dagger gate (adjoint of T gate)
    def tdg(self, qubit):
        TDGGate = Gate('tdg', 1, None)
        self._append(TDGGate, [qubit], [])
        return
    # Controlled X gate (CNOT)
    def cx(self, ctrl_qubit, trgt_qubit):
        CXGate = Gate('cx', 2, None)
        self._append(CXGate, [ctrl_qubit, trgt_qubit], [])
        return
    # Controlled Z gate (CZ)
    def cz(self, ctrl_qubit, trgt_qubit):
        CZGate = Gate('cz', 2, None)
        self._append(CZGate, [ctrl_qubit, trgt_qubit], [])
        return
    # Toffoli gate
    def toffoli(self, ctrl_qubit_1, ctrl_qubit_2, trgt_qubit):
        TOFFGate = Gate('toffoli', 3, None)
        self._append(TOFFGate, [ctrl_qubit_1, ctrl_qubit_2, trgt_qubit], [])
        return
    # Measure qubits in array 'qubits' and store classical outcome in 'cbits'
    def measure(self, qubits, cbits):
        assert(len(qubits) == len(cbits))
        Measure = Gate('measure', len(qubits), None)
        self._append(Measure, qubits, cbits)
        return


####################
# In-House Quantum Logic
####################

def AND(qc, in_qubit_1, in_qubit_2, out_qubit):
    #  AND gate
    #  ALREADY IMPLEMENTED
    qc.toffoli(in_qubit_1, in_qubit_2, out_qubit)
    return

def OR(qc, in_qubit_1, in_qubit_2, out_qubit):
    #  OR gate
    #  YOUR IMPLEMENTATION HERE


    return

def nAND(qc, in_qubits, out_qubit, ancilla):
    #  n-bit AND gate
    #  YOUR IMPLEMENTATION HERE


    return

def nOR(qc, in_qubits, out_qubit, ancilla):
    #  n-bit OR gate
    #  YOUR IMPLEMENTATION HERE


    return

def MAJ(qc, in_qubit_1, in_qubit_2, in_qubit_3, out_qubit):
    #  MAJ gate
    #  YOUR IMPLEMENTATION HERE


    return


def nMAJ(qc, in_qubits, out_qubit, ancilla):
    #  n-bit MAJ gate
    #  YOUR IMPLEMENTATION HERE


    return


####################
# Test functions (feel free to add more of your own tests here)
####################
def testAND():
    print("Testing AND gate...")
    qc = QuantumCircuit(3, 3)
    # Initialize input to 1/sqrt(2)(|000> + |110>)
    qc.h(0)
    qc.cx(0,1)
    # Perform AND
    AND(qc, 0, 1, 2)
    # Examine circuit
    print(qc)
    # Expected output: 1/sqrt(2)(|000> + |111>)

def testOR():
    print("Testing OR gate...")
    qc = QuantumCircuit(3, 3)
    OR(qc, 0, 1, 2)
    print(qc)

def testnAND():
    print("Testing nAND gate...")
    qc = QuantumCircuit(9, 9)
    nAND(qc, [0,1,2,3], 4, [5,6,7,8])
    print(qc)

def testnOR():
    print("Testing nOR gate...")
    qc = QuantumCircuit(9, 9)
    nOR(qc, [0,1,2,3], 4, [5,6,7,8])
    print(qc)

def testMAJ():
    print("Testing MAJ gate...")
    qc = QuantumCircuit(4, 4)
    MAJ(qc, 0, 1, 2, 3)
    print(qc)

def testnMAJ():
    print("Testing nMAJ gate...")
    qc = QuantumCircuit(11, 11)
    nMAJ(qc, [0,1,2,3,4], 5, [6,7,8,9,10])
    print(qc)


####################
# Main tests
####################

def testAll():
    testAND()
    testOR()
    testnAND()
    testnOR()
    # testMAJ()
    # testnMAJ()
    # ... You can add more or comment out tests

def main():
    requirement_A3.check()
    testAll()

if __name__ == '__main__':
    main()
